﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.ConditionActionEncoding;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Selection;

namespace HeuristicLab.Algorithms.LearningClassifierSystems {
  /// <summary>
  /// An operator which represents the main loop of a learning classifier system.
  /// </summary>
  [Item("LearningClassifierSystemMainLoop", "An operator which represents the main loop of a learning classifier system.")]
  [StorableClass]
  public sealed class LearningClassifierSystemMainLoop : AlgorithmOperator {

    #region Parameter Properties
    public IConstrainedValueParameter<ISelector> SelectorParameter {
      get { return (IConstrainedValueParameter<ISelector>)Parameters["Selector"]; }
    }
    public IConstrainedValueParameter<ICrossover> CrossoverParameter {
      get { return (IConstrainedValueParameter<ICrossover>)Parameters["Crossover"]; }
    }
    public IConstrainedValueParameter<IManipulator> MutatorParameter {
      get { return (IConstrainedValueParameter<IManipulator>)Parameters["Mutator"]; }
    }
    #endregion

    #region private properties
    private SolutionsCreator initialSolutionsCreator;
    private MatchConditionOperator matchConditionOperator;
    private PredictionArrayCalculator predictionArrayCalculator;
    private ActionSelector actionSelector;
    private CoveringOperator covering;

    private Placeholder evaluator;
    private Placeholder actionExecuter;
    private Placeholder classifierFetcher;
    #endregion

    [StorableConstructor]
    private LearningClassifierSystemMainLoop(bool deserializing) : base(deserializing) { }
    private LearningClassifierSystemMainLoop(LearningClassifierSystemMainLoop original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new LearningClassifierSystemMainLoop(this, cloner);
    }
    public LearningClassifierSystemMainLoop()
      : base() {
      Initialize();
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ConstrainedValueParameter<ISelector>("Selector", "The operator used to select solutions for reproduction.", new ItemSet<ISelector>() { new ProportionalSelector() }, new ProportionalSelector()));
      Parameters.Add(new ConstrainedValueParameter<ICrossover>("Crossover", "The operator used to cross solutions.", new ItemSet<ICrossover>() { new HeuristicLab.Encodings.CombinedIntegerVectorEncoding.SinglePointCrossover() }, new HeuristicLab.Encodings.CombinedIntegerVectorEncoding.SinglePointCrossover()));
      Parameters.Add(new ValueLookupParameter<IOperator>("Evaluator", "The operator used to evaluate solutions. This operator is executed in parallel, if an engine is used which supports parallelization."));
      UniformOnePositionManipulator test = new UniformOnePositionManipulator();
      test.IntegerVectorParameter.ActualName = "CombinedIntegerVector";
      Parameters.Add(new OptionalConstrainedValueParameter<IManipulator>("Mutator", "The operator used to mutate solutions.", new ItemSet<IManipulator>() { new UniformOnePositionManipulator() }, test));
      #endregion

      #region Create operators
      VariableCreator variableCreator = new VariableCreator();
      ConditionalBranch initialPopulationConditionalBranch = new ConditionalBranch();
      initialSolutionsCreator = new SolutionsCreator();
      IntCounter iterationCounter = new IntCounter();
      UniformSubScopesProcessor matchCondtionSubScopesProcessor = new UniformSubScopesProcessor();
      matchConditionOperator = new MatchConditionOperator();
      predictionArrayCalculator = new PredictionArrayCalculator();
      actionSelector = new ActionSelector();
      UniformSubScopesProcessor matchActionSubScopesProcessor = new UniformSubScopesProcessor();
      MatchActionOperator matchActionOperator = new MatchActionOperator();
      ConditionalSelector conditionMatchSelector = new ConditionalSelector();
      ConditionalSelector actionMatchSelector = new ConditionalSelector();
      SubScopesProcessor matchSetSubScopesProcessor = new SubScopesProcessor();
      CountNumberOfUniqueActions countNumberOfUniqueActions = new CountNumberOfUniqueActions();
      covering = new CoveringOperator();
      MergingReducer mergingReducer = new MergingReducer();
      MergingReducer mergingReducerWithSuccessor = new MergingReducer();
      evaluator = new Placeholder();
      SubScopesProcessor actionSetSubScopesProcessor = new SubScopesProcessor();
      DataReducer actionSetSizeDataReducer = new DataReducer();
      UniformSubScopesProcessor accuracySubScopesProcessor = new UniformSubScopesProcessor();
      CalculateAccuracy calculateAccuracy = new CalculateAccuracy();
      SumAccuracy sumAccuracy = new SumAccuracy();
      UniformSubScopesProcessor updateParametersSubScopesProcessor = new UniformSubScopesProcessor();
      LCSAdaptedGeneticAlgorithm adaptedGeneticAlgorithmMainLoop = new LCSAdaptedGeneticAlgorithm();

      classifierFetcher = new Placeholder();
      actionExecuter = new Placeholder();

      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("ZeroIntValue", new IntValue(0)));
      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("OneIntValue", new IntValue(1)));
      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("Iteration", new IntValue(0)));

      initialPopulationConditionalBranch.ConditionParameter.ActualName = "CreateInitialPopulation";

      initialSolutionsCreator.NumberOfSolutionsParameter.ActualName = "N";

      iterationCounter.ValueParameter.ActualName = "Iteration";
      iterationCounter.IncrementParameter.ActualName = "OneIntValue";

      matchCondtionSubScopesProcessor.Operator = matchConditionOperator;

      matchConditionOperator.MatchParameter.ActualName = "CombinedIntegerVector";

      conditionMatchSelector.CopySelected = new BoolValue(false);
      conditionMatchSelector.ConditionParameter.ActualName = "MatchCondition";

      countNumberOfUniqueActions.ClassifiersParameter.ActualName = "CombinedIntegerVector";

      covering.ActionsInMatchSetParameter.ActualName = countNumberOfUniqueActions.UniqueActionsParameter.ActualName;
      covering.ParallelParameter.Value.Value = true;
      covering.RandomParameter.ActualName = "Random";

      matchActionSubScopesProcessor.Operator = matchActionOperator;

      matchActionOperator.MatchParameter.ActualName = "CombinedIntegerVector";
      matchActionOperator.TargetMatchParameter.ActualName = actionSelector.SelectedActionParameter.ActualName;

      predictionArrayCalculator.MatchParameter.ActualName = "CombinedIntegerVector";

      actionSelector.PredictionArrayParameter.ActualName = predictionArrayCalculator.PredictionArrayParameter.Name;
      actionSelector.RandomParameter.ActualName = "Random";
      actionSelector.ExplorationProbabilityParameter.ActualName = "ExplorationProbability";

      actionMatchSelector.CopySelected = new BoolValue(false);
      actionMatchSelector.ConditionParameter.ActualName = "MatchAction";

      SelectorParameter.Value.CopySelected = new BoolValue(true);
      SelectorParameter.Value.NumberOfSelectedSubScopesParameter.Value = new IntValue(200);

      evaluator.Name = "Evaluator";

      classifierFetcher.Name = "ClassifierFetcher";

      actionExecuter.Name = "ActionExecuter";

      actionSetSizeDataReducer.TargetParameter.ActualName = "CurrentActionSetSize";
      actionSetSizeDataReducer.ParameterToReduce.ActualName = "Numerosity";
      actionSetSizeDataReducer.ReductionOperation.Value = new ReductionOperation(ReductionOperations.Sum);
      actionSetSizeDataReducer.TargetOperation.Value = new ReductionOperation(ReductionOperations.Assign);

      calculateAccuracy.AlphaParameter.ActualName = "Alpha";
      calculateAccuracy.ErrorParameter.ActualName = "Error";
      calculateAccuracy.ErrorZeroParameter.ActualName = "ErrorZero";
      calculateAccuracy.PowerParameter.ActualName = "v";

      sumAccuracy.AccuracyParameter.ActualName = calculateAccuracy.AccuracyParameter.ActualName;
      sumAccuracy.NumerosityParameter.ActualName = "Numerosity";

      adaptedGeneticAlgorithmMainLoop.SelectorParameter.ActualName = SelectorParameter.Name;
      adaptedGeneticAlgorithmMainLoop.CrossoverParameter.ActualName = CrossoverParameter.Name;
      adaptedGeneticAlgorithmMainLoop.MutatorParameter.ActualName = MutatorParameter.Name;
      adaptedGeneticAlgorithmMainLoop.RandomParameter.ActualName = "Random";
      adaptedGeneticAlgorithmMainLoop.MaximumGenerationsParameter.ActualName = "ZeroIntValue";
      adaptedGeneticAlgorithmMainLoop.QualityParameter.ActualName = "Fitness";
      //needed?
      adaptedGeneticAlgorithmMainLoop.MutationProbabilityParameter.Value = new PercentValue(10);
      adaptedGeneticAlgorithmMainLoop.MaximizationParameter.Value = new BoolValue(true);
      #endregion

      #region Create operator graph
      OperatorGraph.InitialOperator = variableCreator;

      variableCreator.Successor = initialPopulationConditionalBranch;
      initialPopulationConditionalBranch.TrueBranch = initialSolutionsCreator;
      initialPopulationConditionalBranch.FalseBranch = new EmptyOperator();
      initialPopulationConditionalBranch.Successor = classifierFetcher;
      classifierFetcher.Successor = matchCondtionSubScopesProcessor;
      matchCondtionSubScopesProcessor.Successor = conditionMatchSelector;
      //variableCreator.Successor = adaptedGeneticAlgorithmMainLoop;
      //adaptedGeneticAlgorithmMainLoop.Successor = conditionMatchSelector;

      conditionMatchSelector.Successor = matchSetSubScopesProcessor;
      matchSetSubScopesProcessor.Operators.Add(new EmptyOperator());
      matchSetSubScopesProcessor.Operators.Add(countNumberOfUniqueActions);
      matchSetSubScopesProcessor.Successor = mergingReducerWithSuccessor;
      countNumberOfUniqueActions.Successor = covering;
      covering.Successor = predictionArrayCalculator;
      predictionArrayCalculator.Successor = actionSelector;
      actionSelector.Successor = matchActionSubScopesProcessor;
      matchActionSubScopesProcessor.Successor = actionMatchSelector;
      actionMatchSelector.Successor = actionExecuter;
      actionExecuter.Successor = actionSetSubScopesProcessor;
      actionSetSubScopesProcessor.Operators.Add(new EmptyOperator());
      actionSetSubScopesProcessor.Operators.Add(actionSetSizeDataReducer);
      actionSetSizeDataReducer.Successor = accuracySubScopesProcessor;
      accuracySubScopesProcessor.Operator = calculateAccuracy;
      accuracySubScopesProcessor.Successor = sumAccuracy;
      sumAccuracy.Successor = updateParametersSubScopesProcessor;
      updateParametersSubScopesProcessor.Operator = evaluator;

      actionSetSubScopesProcessor.Successor = mergingReducer;

      mergingReducerWithSuccessor.Successor = iterationCounter;
      iterationCounter.Successor = classifierFetcher;
      //mergingReducer.Successor = adaptedGeneticAlgorithmMainLoop;
      #endregion
    }

    private void ParameterizeStochasticOperator(IOperator op) {
      IStochasticOperator stochasticOp = op as IStochasticOperator;
      if (stochasticOp != null) {
        stochasticOp.RandomParameter.ActualName = "Random";
        stochasticOp.RandomParameter.Hidden = true;
      }
    }

    internal void SetCurrentProblem(IConditionActionProblem problem) {
      initialSolutionsCreator.SolutionCreatorParameter.ActualName = problem.SolutionCreatorParameter.Name;
      initialSolutionsCreator.EvaluatorParameter.ActualName = problem.EvaluatorParameter.Name;

      problem.ActionExecuter.SelectedActionParameter.ActualName = actionSelector.SelectedActionParameter.ActualName;

      problem.ClassifierFetcher.IterationParameter.ActualName = "Iteration";

      evaluator.OperatorParameter.ActualName = problem.EvaluatorParameter.Name;

      classifierFetcher.OperatorParameter.ActualName = problem.ClassifierFetcherParameter.Name;

      actionExecuter.OperatorParameter.ActualName = problem.ActionExecuterParameter.Name;

      matchConditionOperator.TargetMatchParameter.ActualName = problem.ClassifierFetcher.CurrentClassifierToMatchParameter.ActualName;

      covering.SolutionCreatorParameter.ActualName = problem.CoveringSolutionCreatorParameter.Name;
      covering.EvaluatorParameter.ActualName = problem.EvaluatorParameter.Name;
      covering.MinimalNumberOfUniqueActionsParameter.ActualName = problem.ThetaMinimalNumberOfActionsParameter.Name;
      covering.PossibleActionsParameter.ActualName = problem.PossibleActionsParameter.Name;

      predictionArrayCalculator.PredictionParameter.ActualName = problem.Evaluator.PredictionParameter.ActualName;
      predictionArrayCalculator.FitnessParameter.ActualName = problem.Evaluator.FitnessParameter.ActualName;
    }
    //private void ParameterizeSelectors() {
    //  foreach (ISelector selector in SelectorParameter.ValidValues) {
    //    selector.CopySelected = new BoolValue(true);
    //    //set value by parameter!
    //    selector.NumberOfSelectedSubScopesParameter.Value = new IntValue(5);
    //    selector.NumberOfSelectedSubScopesParameter.Hidden = true;
    //    ParameterizeStochasticOperator(selector);
    //  }
    //  if (Problem != null) {
    //    foreach (ISingleObjectiveSelector selector in SelectorParameter.ValidValues.OfType<ISingleObjectiveSelector>()) {
    //      selector.MaximizationParameter.Value = new BoolValue(true);
    //      selector.MaximizationParameter.Hidden = true;
    //      selector.QualityParameter.ActualName = Problem.Evaluator.QualityParameter.ActualName;
    //      selector.QualityParameter.Hidden = true;
    //    }
    //  }
    //}
  }
}
